import os
from dataclasses import dataclass, field
from typing import Dict, List, List, Any, TypeVar, Sequence
from dbt.contracts.graph.parsed import ParsedModelNode
from dbt.node_types import NodeType
from pathlib import Path
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.results import RunResultsArtifact, RunResultOutput


class FalGeneralException(Exception):
    pass


@dataclass
class DbtModel:
    node: ParsedModelNode
    name: str = field(init=False)
    meta: Dict[str, Any] = field(init=False)
    status: str = field(init=False)
    columns: Dict[str, Any] = field(init=False)

    def __post_init__(self):
        self.name = self.node.name
        self.meta = self.node.config.meta
        self.columns = self.node.columns
        self.unique_id = self.node.unique_id

    def __hash__(self) -> int:
        return self.unique_id.__hash__()        

    def get_scipts(self, keyword, dbt_root) -> List[Path]:
        scripts = self.node.config.meta[keyword]["scripts"]
        return list(map(lambda script: Path(os.path.realpath(os.path.normpath(dbt_root + "/" + script))), scripts))

@dataclass
class DbtManifest:
    nativeManifest: Manifest

    def get_models(self) -> List[DbtModel]:
        return list(
            filter(
                lambda model: model.node.resource_type == NodeType.Model,
                map(
                    lambda node: DbtModel(node=node), self.nativeManifest.nodes.values()
                ),
            )
        )


@dataclass
class DbtRunResult:
    nativeRunResult: RunResultsArtifact
    results: Sequence[RunResultOutput] = field(init=False)

    def __post_init__(self):
        self.results = self.nativeRunResult.results


T = TypeVar("T", bound="DbtProject")


@dataclass
class DbtProject:
    name: str
    model_config_paths: List[str]
    models: List[DbtModel]
    manifest: DbtManifest
    keyword: str
    scripts: List[Path]
    run_result: DbtRunResult

    def state_has_changed(self, other: DbtManifest) -> bool:
        return self.manifest != other

    def find_model_location(self, model: DbtModel) -> List[str]:
        model_node = self.manifest.nodes[model.unique_id]
        return model_node.relation_name.replace("`", "")

    def changed_models_ids(self) -> List[str]:
        return list(map(lambda res: res["unique_id"], self.run_result.results))

    def get_models_with_keyword(self, keyword) -> List[DbtModel]:
        return list(filter(lambda model: keyword in model.meta, self.models))

    def get_filtered_models(self, all):
        changed_models_ids = self.changed_models_ids()
        filtered_models: List[DbtModel] = []

        for node in self.get_models_with_keyword(self.keyword):
            if all:
                filtered_models.append(node)
            elif node.unique_id in changed_models_ids:
                filtered_models.append(node)

        return filtered_models



import click
import os

from dbt.logger import log_manager
from dbt.config.profile import DEFAULT_PROFILES_DIR

from fal.run_scripts import run_scripts
from faldbt.parse import parse_project
import faldbt.lib as lib


@click.group()
@click.version_option()
def cli():
    pass


@cli.command()
@click.option(
    "--project-dir",
    default=os.getcwd(),
    help="Directory to look for dbt_project.yml.",
    type=click.Path(exists=True),
)
@click.option(
    "--profiles-dir",
    default=DEFAULT_PROFILES_DIR,
    help="Directory to look for profiles.yml.",
    type=click.Path(exists=True),
)
@click.option(
    "--keyword",
    default="fal",
    help="Property in meta to look for fal configurations.",
    type=click.STRING,
)
@click.option(
    "--all",
    help="Only run models that ran in the last dbt run.",
    is_flag=True,
)
@click.option(
    "--debug",
    help="Display debug logging during execution.",
    is_flag=True,
)
def run(project_dir, profiles_dir, keyword, all, debug):
    with log_manager.applicationbound():
        if debug:
            log_manager.set_debug()

        real_project_dir = os.path.realpath(os.path.normpath(project_dir))
        real_profiles_dir = os.path.realpath(os.path.normpath(profiles_dir))

        project = parse_project(real_project_dir, real_profiles_dir, keyword)
        models = project.get_filtered_models(all)
        for model in models:
            run_scripts(
                model, keyword, project.manifest.nativeManifest, real_project_dir
            )



from dataclasses import dataclass, field
from os import path
from typing import List,TypeVar, Dict
from faldbt.project import DbtModel 
from pathlib import Path
import re

T = TypeVar("T", bound="FalScript")

@dataclass(frozen=True)
class FalScript:
    model: DbtModel
    path: str
    dependencies: List[T] = field(default_factory=list)

    def __hash__(self):
        return self.path.__hash__() * self.model.name.__hash__()

    def __repr__(self):
        return "FalScript(" + self.path.__str__() + ")"

    def parse(self) -> List[str]:
        """
        Parses the python file to identify any dependencies
        """
        with open(self.path) as file:
            lines = file.readlines()
            ## TODO: do this search in parallel
            dependency_names = []
            for line in lines:
                match = re.search('ref\("(.*)"\)', line)
                if match:
                    group = match.group(1)
                    dependency_names.append(group)
            return dependency_names    
    
    def add_dependency(self, dependency: T):
        if (dependency not in self.dependencies):
            self.dependencies.append(dependency)

    def exec(self, ref, context, source):
        """
        Executes the script
        """
        with open(path) as file:
            a_script = file.read()
            exec(
                a_script,
                {
                    "ref": ref,
                    "context": context,
                    "source": source,
                },
            )

@dataclass(frozen=True)
class UniqueKey:
    model: DbtModel
    script_path: Path

class ScriptGraphBuilder:
    modelNameToModelLookup: Dict[str, DbtModel] = {}
    modelToScriptLookup: Dict[str, List[Path]] = {}
    falScripts: Dict[UniqueKey, FalScript] = {}

    def __init__(self, models: List[DbtModel], keyword: str, root: str):
        for model in models:
            self.modelToScriptLookup[model.name] = model.get_scipts(keyword, root)
            self.modelNameToModelLookup[model.name] = model
        for model_name, script_list in self.modelToScriptLookup.items():
            for script in script_list:
                self.recursively_set_dependencies(UniqueKey(self.modelNameToModelLookup[model_name], script))

    def recursively_set_dependencies(self, key: UniqueKey):
        if key in self.falScripts:
          return
        self.falScripts[key] = FalScript(key.model, key.script_path)
        current_script = self.falScripts.get(key)
        dependency_models = current_script.parse()
       
        for model_name in dependency_models:
            model = self.modelNameToModelLookup[model_name]
            dependency_scripts = self.modelToScriptLookup.get(model_name, [])

            for dependency_script in dependency_scripts:
                dependency_key = UniqueKey(model, dependency_script)
                self.recursively_set_dependencies(dependency_key)
                current_script.add_dependency(self.falScripts.get(dependency_key))
    
    def get_values(self) -> List[FalScript]:
        return self.falScripts.values()


class ScriptGraph:
    graph: List[FalScript]
    outgoing: Dict[FalScript, List[FalScript]] = {}
    incoming: Dict[FalScript, List[FalScript]] = {}
    ordered_list: List[FalScript] = []

    def __init__(self, models: List[DbtModel], keyword: str, root: str):
        self.graph = ScriptGraphBuilder(models, keyword, root).get_values()
        self.incoming = dict(map(lambda script: [script, script.dependencies], self.graph))
        for edge in self.graph:
            for dependency in edge.dependencies:
                if (self.outgoing.get(dependency) is None):
                    self.outgoing[dependency] = [] 
                self.outgoing[dependency].append(edge)
            if (self.outgoing.get(edge) is None):
                self.outgoing[edge] = []
        

    def sort(self):
        """
        Topologically sorts the python scripts to make sure all dependencies
        run before the script itsef
        """
        leaf_nodes = list(filter(lambda key: len(self.incoming[key]) == 0, self.incoming.keys()))
        while leaf_nodes:
            node = leaf_nodes.pop(0)
            self.ordered_list.append(node)
            outgoing_copy = self.outgoing[node].copy()
            for item in outgoing_copy:
                self.incoming[item].remove(node)
                self.outgoing[node].remove(item)
                if (not self.incoming[item]):
                    leaf_nodes.append(item)
        detect_cycles = _flatten(self.outgoing.values())
        if (detect_cycles):
            raise Exception("Your python scripts contain a cycle could not determine the right order")
        return self.ordered_list


def _flatten(t):
    return [item for sublist in t for item in sublist]



import os
import json
import glob
from typing import Dict, Any

import dbt.tracking
from dbt.contracts.results import RunResultsArtifact

import faldbt.lib as lib
from faldbt.utils.yaml_helper import load_yaml_text
from faldbt.project import DbtProject, DbtManifest, DbtRunResult


class FalParseError(Exception):
    pass


def _load_file_contents(path: str, strip: bool = True) -> str:
    with open(path, "rb") as handle:
        to_return = handle.read().decode("utf-8")

    if strip:
        to_return = to_return.strip()

    return to_return


def _read_yaml(path):
    contents = _load_file_contents(path)
    return load_yaml_text(contents)


def _read_json(path: str) -> Dict[str, Any]:
    return json.loads(_load_file_contents(path))


def _flatten(t):
    return [item for sublist in t for item in sublist]


def _get_all_model_config(project_root, project_dict):
    return _flatten(
        map(
            ## find one with any kind yml this doesnt need to schema
            ## look at all of them find the ones that has model in them
            ## and keep remembering it
            lambda model_path: glob.glob(
                os.path.join(project_root, model_path, "**.yml"),
                recursive=True,
            ),
            project_dict["source-paths"],
        )
    )


def parse_project(project_dir: str, profiles_dir: str, keyword: str):
    project_dict = _get_project_dict(project_dir)
    scripts = glob.glob(os.path.join(project_dir, "**.py"), recursive=True)
    model_config_paths = _get_all_model_config(project_dir, project_dict)
    target_path = os.path.join(project_dir, project_dict["target-path"])

    run_results_path = os.path.join(target_path, "run_results.json")
    try:
        run_results = _read_json(run_results_path)
    except IOError as e:
        raise FalParseError("Did you forget to run dbt run?") from e

    config = lib.get_dbt_config(project_dir)
    lib.register_adapters(config)

    # Necessary for parse_to_manifest to not fail
    dbt.tracking.initialize_tracking(profiles_dir)

    manifest = lib.parse_to_manifest(config)
    run_result_artifact = RunResultsArtifact(**run_results)
    dbtmanifest = DbtManifest(nativeManifest=manifest)

    models = dbtmanifest.get_models()
    status_map = dict(
        map(lambda result: [result["unique_id"], result["status"]], run_result_artifact)
    )
    for model in models:
        model.status = status_map.get(model.unique_id)

    return DbtProject(
        name=project_dict["name"],
        model_config_paths=list(model_config_paths),
        models=models,
        manifest=DbtManifest(nativeManifest=manifest),
        keyword=keyword,
        scripts=scripts,
        run_result=DbtRunResult(run_result_artifact),
    )


def _get_project_dict(project_dir):
    project_yaml_filepath = os.path.join(project_dir, "dbt_project.yml")

    if not os.path.exists(project_yaml_filepath):
        raise FalParseError(
            "no dbt_project.yml found at expected path {}".format(project_yaml_filepath)
        )

    project_dict = _read_yaml(project_yaml_filepath)

    if not isinstance(project_dict, dict):
        raise FalParseError("dbt_project.yml formatting is wrong")

    return project_dict



# NOTE: COPIED FROM https://github.com/dbt-labs/dbt-core/blob/40ae6b6bc860a30fa383756b7cdb63709ce829a8/core/dbt/lib.py
import os
import six

from datetime import datetime
from uuid import uuid4
from collections import namedtuple
from typing import List, Tuple, Type, Union

from dbt.config.runtime import RuntimeConfig
from dbt.contracts.connection import AdapterResponse
from dbt.contracts.graph.manifest import Manifest
import dbt.clients.agate_helper as agate_helper
import dbt.adapters.factory as adapters_factory

import pandas as pd

from sqlalchemy.sql.ddl import CreateTable
from sqlalchemy.sql import Insert

from faldbt.cp.contracts.graph.parsed import ParsedModelNode, ParsedSourceDefinition
from faldbt.cp.contracts.sql import ResultTable, RemoteRunResult

import agatesql
import agatesql.table

# from sqlalchemy.engine import Connection as SQLAlchemyConnection

RuntimeArgs = namedtuple("RuntimeArgs", "project_dir profiles_dir single_threaded")


def get_dbt_config(project_dir: str, single_threaded=False):
    from dbt.config.runtime import RuntimeConfig

    if os.getenv("DBT_PROFILES_DIR"):
        profiles_dir = os.getenv("DBT_PROFILES_DIR")
    else:
        profiles_dir = os.path.expanduser("~/.dbt")

    # Construct a phony config
    return RuntimeConfig.from_args(
        RuntimeArgs(project_dir, profiles_dir, single_threaded)
    )


def register_adapters(config: RuntimeConfig):
    # Clear previously registered adapters. This fixes cacheing behavior on the dbt-server
    adapters_factory.reset_adapters()
    # Load the relevant adapter
    adapters_factory.register_adapter(config)


def _get_operation_node(manifest: Manifest, project_path, sql):
    from dbt.parser.manifest import process_node
    from faldbt.cp.parser.sql import SqlBlockParser

    config = get_dbt_config(project_path)
    block_parser = SqlBlockParser(
        project=config,
        manifest=manifest,
        root_project=config,
    )

    # NOTE: nodes get registered to the manifest automatically,
    # HACK: we need to include uniqueness (UUID4) to avoid clashes
    name = "SQL:" + str(hash(sql)) + ":" + str(uuid4())
    sql_node = block_parser.parse_remote(sql, name)
    process_node(config, manifest, sql_node)
    return sql_node


# NOTE: Once we get an adapter, we must call `connection_for` or `connection_named` to use it
def _get_adapter(project_path: str):
    config = get_dbt_config(project_path)

    adapters_factory.cleanup_connections()
    return adapters_factory.get_adapter(config)


def _execute_sql(
    manifest: Manifest, project_path: str, sql: str
) -> Tuple[AdapterResponse, RemoteRunResult]:
    node = _get_operation_node(manifest, project_path, sql)
    adapter = _get_adapter(project_path)

    result = None
    with adapter.connection_for(node):
        adapter.connections.begin()
        response, execute_result = adapter.execute(sql, fetch=True)

        table = ResultTable(
            column_names=list(execute_result.column_names),
            rows=[list(row) for row in execute_result],
        )

        result = RemoteRunResult(
            raw_sql=sql,
            compiled_sql=sql,
            node=node,
            table=table,
            timing=[],
            logs=[],
            generated_at=datetime.utcnow(),
        )
        adapter.connections.commit()

    return response, result


def _get_target_relation(
    target: Union[ParsedModelNode, ParsedSourceDefinition], project_path: str
):
    adapter = _get_adapter(project_path)

    relation = None
    with adapter.connection_named(str(uuid4())):
        # This ROLLBACKs so it has to be a new connection
        relation = adapter.get_relation(
            target.database, target.schema, target.identifier
        )
    return relation


def execute_sql(manifest: Manifest, project_path: str, sql: str) -> RemoteRunResult:
    _, result = _execute_sql(manifest, project_path, sql)
    return result


def fetch_target(
    manifest: Manifest,
    project_path: str,
    target: Union[ParsedModelNode, ParsedSourceDefinition],
) -> RemoteRunResult:
    relation = _get_target_relation(target, project_path)

    if relation is None:
        raise Exception(f"Could not get relation for '{target.unique_id}'")

    query = f"SELECT * FROM {relation}"
    _, result = _execute_sql(manifest, project_path, query)
    return result


def write_target(
    data: pd.DataFrame,
    manifest: Manifest,
    project_path: str,
    target: Union[ParsedModelNode, ParsedSourceDefinition],
) -> RemoteRunResult:
    relation = _get_target_relation(target, project_path)

    column_names: List[str] = list(data.columns)
    rows = data.to_records(index=False)
    row_dicts = list(map(lambda row: dict(zip(column_names, row)), rows))

    agate_table = agate_helper.table_from_data(row_dicts, column_names=column_names)

    # We are using the SQLAlchemy table to generate the SQL, but it targeting sqlite
    # This may bite us later for other adapters
    alchemy_table = agatesql.table.make_sql_table(
        agate_table, table_name=target.identifier, db_schema=target.schema
    )

    if relation is None:
        create_stmt = CreateTable(alchemy_table).compile(
            compile_kwargs={"literal_binds": True}
        )

        _execute_sql(manifest, project_path, six.text_type(create_stmt).strip())

    insert_stmt = (
        Insert(alchemy_table)
        .values(row_dicts)
        .compile(compile_kwargs={"literal_binds": True})
    )

    _, result = _execute_sql(manifest, project_path, six.text_type(insert_stmt).strip())
    return result


def parse_to_manifest(config):
    from dbt.parser.manifest import ManifestLoader

    return ManifestLoader.get_full_manifest(config)



import os
from typing import Optional, Union, Dict, Any

import faldbt.lib as lib

from dbt.contracts.graph.manifest import Manifest, MaybeNonSource, MaybeParsedSource
from dbt.contracts.graph.parsed import ParsedModelNode
from dbt.contracts.results import RunStatus, TestStatus, FreshnessStatus

from faldbt.cp.contracts.graph.parsed import ColumnInfo
from faldbt.project import DbtModel

import pandas as pd
from dataclasses import dataclass


@dataclass
class CurrentModel:
    name: str
    status: Union[RunStatus, TestStatus, FreshnessStatus]
    columns: Dict[str, ColumnInfo]
    meta: Dict[Any, Any]


@dataclass
class Context:
    current_model: CurrentModel


def run_scripts(model: DbtModel, keyword: str, manifest: Manifest, dbt_dir: str):
    for script in model.meta.get(keyword, {}).get("scripts", []):
        meta = model.meta
        _del_key(meta, keyword)

        current_model = CurrentModel(
            name=model.name, status=model.status, columns=model.columns, meta=meta
        )

        context = Context(current_model=current_model)
        real_script = os.path.join(dbt_dir, script)
        with open(real_script) as file:
            a_script = file.read()
            exec(
                a_script,
                {
                    "ref": _get_ref_resolver(model.node, manifest, dbt_dir),
                    "context": context,
                    "source": _get_source_resolver(model.node, manifest, dbt_dir),
                    "write_to_source": _write_to_source(model.node, manifest, dbt_dir),
                },
            )


def _del_key(dict: Dict[str, Any], key: str):
    try:
        del dict[key]
    except KeyError:
        pass


def _get_ref_resolver(
    model: ParsedModelNode,
    manifest: Manifest,
    dbt_dir: str,
):
    def ref_resolver(target_model_name: str, target_package_name: Optional[str] = None):
        target_model: MaybeNonSource = manifest.resolve_ref(
            target_model_name, target_package_name, dbt_dir, model.package_name
        )
        if target_model is None:
            raise Exception(
                f"Could not find model {target_model_name}{target_package_name or ''}"
            )

        result = lib.fetch_target(manifest, dbt_dir, target_model)
        return pd.DataFrame.from_records(
            result.table.rows, columns=result.table.column_names
        )

    return ref_resolver


def _get_source_resolver(
    model: ParsedModelNode,
    manifest: Manifest,
    dbt_dir: str,
):
    def source_resolver(target_source_name: str, target_table_name: str):
        target_source: MaybeParsedSource = manifest.resolve_source(
            target_source_name, target_table_name, dbt_dir, model.package_name
        )

        if target_source is None:
            raise Exception(
                f"Could not find source {target_source_name}.{target_table_name}"
            )
        result = lib.fetch_target(manifest, dbt_dir, target_source)
        return pd.DataFrame.from_records(
            result.table.rows, columns=result.table.column_names
        )

    return source_resolver


def _write_to_source(
    model: ParsedModelNode,
    manifest: Manifest,
    dbt_dir: str,
):
    def source_writer(
        data: pd.DataFrame, target_source_name: str, target_table_name: str
    ):
        target_source: MaybeParsedSource = manifest.resolve_source(
            target_source_name, target_table_name, dbt_dir, model.package_name
        )

        if target_source is None:
            raise Exception(
                f"Could not find source '{target_source_name}'.'{target_table_name}'"
            )

        lib.write_target(data, manifest, dbt_dir, target_source)

    return source_writer
